;;; DO NOT MODIFY THIS FILE
(if (featurep 'escreen-autoloads) (error "Already loaded"))

;;;### (autoloads nil "_pkg" "escreen/_pkg.el")

(package-provide 'escreen :version 1.01 :author-version "1.16" :type 'regular)

;;;***

;;;### (autoloads (escreen-install) "escreen" "escreen/escreen.el")

(autoload 'escreen-install "escreen" nil t nil)

;;;***

(provide 'escreen-autoloads)
